import torch
import torch_optimizer as optim

from typing import List, Tuple, Dict, Union, Any
from finol.utils import load_config


class OptimizerSelector:
    """
    Class to select optimizer for portfolio selection.
    """
    def __init__(self, model) -> None:
        self.config = load_config()
        self.model = model

        self.optimizer_dict = {
            "Adadelta": torch.optim.Adadelta,
            "Adagrad": torch.optim.Adagrad,
            "Adam": torch.optim.Adam,
            "AdamW": torch.optim.AdamW,
            # "SparseAdam": torch.optim.SparseAdam,
            "Adamax": torch.optim.Adamax,
            "ASGD": torch.optim.ASGD,
            "SGD": torch.optim.SGD,
            "RAdam": torch.optim.RAdam,
            "Rprop": torch.optim.Rprop,
            "RMSprop": torch.optim.RMSprop,
            "NAdam": torch.optim.NAdam,
            # "LBFGS": torch.optim.LBFGS,
            "A2GradExp": optim.A2GradExp,
            "A2GradInc": optim.A2GradInc,
            "A2GradUni": optim.A2GradUni,
            "AccSGD": optim.AccSGD,
            "AdaBelief": optim.AdaBelief,
            "AdaBound": optim.AdaBound,
            "AdaMod": optim.AdaMod,
            "Adafactor": optim.Adafactor,
            # "Adahessian": optim.Adahessian,
            "AdamP": optim.AdamP,
            "AggMo": optim.AggMo,
            "Apollo": optim.Apollo,
            "DiffGrad": optim.DiffGrad,
            "LARS": optim.LARS,
            "Lamb": optim.Lamb,
            "MADGRAD": optim.MADGRAD,
            "NovoGrad": optim.NovoGrad,
            "PID": optim.PID,
            "QHAdam": optim.QHAdam,
            "QHM": optim.QHM,
            "Ranger": optim.Ranger,
            "RangerQH": optim.RangerQH,
            "RangerVA": optim.RangerVA,
            "SGDP": optim.SGDP,
            "SGDW": optim.SGDW,
            "SWATS": optim.SWATS,
            # "Shampoo": optim.Shampoo,
            "Yogi": optim.Yogi
        }

    def select_optimizer(self, sampled_lr: Union[Dict[str, Any], None] = None, sampled_optimizer: Union[Dict[str, Any], None] = None) -> object:
        """
        Selects an optimizer based on the configuration.

        :param sampled_lr: Sampled learning rate or None. Required only when optimizing the learning rate selection.
        :param sampled_optimizer: Sampled optimizer or None. Required only when optimizing the optimizer selection.
        :return: The selected optimizer for the model.
        :raises ValueError: If the optimizer name is invalid.
        """
        if sampled_lr!= None:
            lr = sampled_lr
        else:
            lr = self.config["LEARNING_RATE"]
        if sampled_optimizer!= None:
            optimizer_name = sampled_optimizer
        else:
            optimizer_name = self.config["OPTIMIZER_NAME"]

        optimizer_cls = self.optimizer_dict.get(optimizer_name, None)
        if optimizer_cls is None:
            raise ValueError(f"Invalid optimizer name: {optimizer_name}. Supported optimizers are: {self.optimizer_dict}")
        return optimizer_cls(self.model.parameters(), lr=lr)


import time
import torch
import numpy as np
import pandas as pd

from typing import List, Tuple, Dict, Union, Any
from finol.utils import load_config, actual_portfolio_selection, add_prefix


class MetricCaculator:
    """
    Class to calculate various performance metrics based on the loaded dataset output and trained model output.
    """
    def __init__(self, load_dataset_output=None, train_model_output=None, mode="normal") -> None:
        self.daily_returns = None
        self.config = load_config()
        self.load_dataset_output = load_dataset_output
        self.train_model_output = train_model_output
        self.mode = mode

        if mode == "normal":
            self.logdir = self.train_model_output["logdir"]
        else:
            self.logdir = None

    def caculate_annual_percentage_yield(self, num_trading_periods: float, final_cumulative_wealth: float) -> float:
        """
        *Annualized Percentage Yield* (APY): Annualize the *Final Cumulative Wealth* based on the investment horizon :math:`n` to facilitate comparison across different datasets.

        .. math::

            \mathrm{APY}=\sqrt[y]{\mathrm{FCW}}-1,

        where :math:`y` is the number of years in the investment period.

        :param num_trading_periods: Number of trading periods :math:`n`.
        :param final_cumulative_wealth: *Final Cumulative Wealth*.
        :return: *Annual Percentage Yield*.
        """
        num_years = num_trading_periods / self.config["NUM_DAYS_PER_YEAR"][self.config["DATASET_NAME"]]  # Number of years
        annual_percentage_yield = pow(final_cumulative_wealth, 1 / num_years) - 1  # APY formula
        return annual_percentage_yield

    def caculate_sharpe_ratio(self, daily_returns: np.ndarray, annual_percentage_yield: float) -> float:
        """
        *Sharpe Ratio* (SR): Measure risk-adjusted return using the portfolio's excess return over risk-free rate per unit of volatility risk.

        .. math::

            \mathrm{SR}=\\frac{\mathrm{APY}-R^f}{\sigma_n},

        where :math:`R^f = 0.04` is the annualized risk-free rate and :math:`\sigma_n` denotes the annualized standard deviation of portfolio daily returns.

        :param daily_returns: Daily return sequence.
        :param annual_percentage_yield: *Annual Percentage Yield*.
        :return: *Sharpe Ratio*.
        """
        sigma = np.std(daily_returns, ddof=1) * pow(self.config["NUM_DAYS_PER_YEAR"][self.config["DATASET_NAME"]], 0.5)  # Standard deviation of daily returns (annualized)
        sharpe_ratio = (annual_percentage_yield - 0.04) / sigma  # Sharpe ratio formula, risk-free rate of return, reference CORN in 2011 is 0.04, reference LOAD in 2019 is 0
        return sharpe_ratio

    def caculate_volatility_risk(self, daily_returns) -> float:
        """
        *Volatility Risk* (VR): Calculate the annualized standard deviation :math:`\sigma_n` of daily portfolio returns, indicating return fluctuation risk.

        :param daily_returns: Daily return sequence.
        :return: *Volatility Risk*.
        """
        sigma = np.std(daily_returns, ddof=1) * pow(self.config["NUM_DAYS_PER_YEAR"][self.config["DATASET_NAME"]], 0.5)  # Standard deviation of daily returns (annualized)
        return sigma

    def caculate_maximum_drawdown(self, daily_cumulative_wealth) -> Tuple[np.ndarray, np.ndarray, float]:
        """
        *Maximum DrawDown* (MDD): Measure the maximum peak-to-trough decline of a portfolio over the investment horizon :math:`n`.

        .. math::

            \mathrm{MDD}=\\max_{0\leq t\leq n}\\left(\\frac{\\max_{0\leq k\leq t} S_k - S_t}{\\max_{0\leq k\leq t} S_k}\\right),

        where :math:`S_t` is the cumulative wealth at the end of the :math:`t`-th trading period.

        :param daily_cumulative_wealth: *Daily Cumulative Wealth*.
        :return: Tuple containing the daily drawdown, daily maximum drawdown, and maximum drawdown.
        """
        daily_drawdown = []
        daily_maximum_drawdown = []
        num_trading_periods = len(daily_cumulative_wealth)
        for i in range(num_trading_periods):
            j = i + 1
            temp = daily_cumulative_wealth[:j]
            ddval = (max(temp) - temp[i]) / max(temp)  # Drawdown value for each trading period
            daily_drawdown.append(ddval)
            daily_maximum_drawdown.append(max(daily_drawdown))
        maximum_drawdown = max(daily_drawdown)  # Maximum drawdown value

        # Convert lists to NumPy arrays
        daily_drawdown = np.array(daily_drawdown)
        daily_maximum_drawdown = np.array(daily_maximum_drawdown)

        return daily_drawdown, daily_maximum_drawdown, maximum_drawdown

    def caculate_average_turnover(self, daily_turnovers) -> None:
        """
        *Average Turnover* (ATO): Calculate the average proportion of portfolio vectors changed between periods, indicating rebalancing frequency.

        .. math::

            \mathrm{ATO}=\\frac{1}{2(n)} \\sum_{t=1}^n\\left\|\mathbf{b}_t-\\hat{\mathbf{b}}_{t-1}\\right\|_1,

        where portfolio :math:`\\hat{\mathbf{b}}_{t-1} = (\\frac{b_{t-1, 1}  x_{t-1, 1}}{\mathbf{b}_{t-1}^{\\top} \mathbf{x}_{t-1}}, \\ldots, \\frac{b_{t-1, m}  x_{t-1, m}}{\mathbf{b}_{t-1}^{\\top} \mathbf{x}_{t-1}})` is the portfolio vector at the end of :math:`(t-1)`-th period adjusted for asset returns. For simplicity, we set all elements of :math:`\hat{\mathbf{b}}_{0}` to 0.

        :param daily_turnovers: Daily turnover sequence.
        :return: *Average Turnover*
        """
        num_trading_periods = sum(daily_turnovers)
        average_turnover = sum(daily_turnovers) / (2 * (num_trading_periods))
        return average_turnover

    def caculate_transaction_costs_wealth(self, portfolios, labels) -> Tuple[np.ndarray, pd.Series]:
        """
        *Transaction Costs-Adjusted Cumulative Wealth* (TCW):  Apply proportional transaction costs to the raw *Cumulative Wealth* equation based on rebalancing.

        .. math::

            \mathrm{TCW} = S_0 \\prod_{t=1}^n\\left[\\left(\mathbf{b}_t^{\\top}  \mathbf{x}_t\\right) \\times\\left(1-\\frac{c}{2} \\times \\left\|\mathbf{b}_t-\\hat{\mathbf{b}}_{t-1}\\right\|_1\\right)\\right],

        where :math:`S_0` is the initial cumulative wealth, :math:`\mathbf{b}_t` is the portfolio vector and :math:`\mathbf{x}_t` is the price relative vector in period :math:`t`. :math:`c \in (0,1)` is the transaction costs rate.

        :param portfolios: Portfolio sequence.
        :param labels: Price relative sequence.
        :return: Tuple containing daily turnover and *transaction costs-adjusted cumulative wealth*.
        """
        num_trading_periods, num_assets = portfolios.shape

        df = pd.DataFrame(columns=["Rate", "TCW"])

        # tc = 0.005
        # interval = 0.001
        for tc_rate in np.arange(0, 0.005 + 0.001, 0.001):
            portfolio_o = np.zeros_like(portfolios[0, :])

            daily_turnover = []
            daily_returns = []

            for i in range(num_trading_periods):
                label = labels[i, :]
                portfolio = portfolios[i, :]

                turnover = (abs(portfolio_o - portfolio).sum())
                daily_return = np.sum(label * portfolio) * (1 - ((tc_rate / 2) * turnover))
                portfolio_o = portfolio * label / daily_return

                daily_turnover.append(turnover)
                daily_returns.append(daily_return)

            df.loc[len(df)] = [tc_rate, np.cumprod(daily_returns)[-1]]

        daily_turnovers = np.array(daily_turnover)
        transaction_costs_wealth = df["TCW"]
        return daily_turnovers, transaction_costs_wealth


    def caculate_final_cumulative_wealth(self, portfolios: np.ndarray, labels: np.ndarray) -> Tuple[np.ndarray, np.ndarray, float]:
        """
        *Final Cumulative Wealth* (FCW): Calculate the final cumulative wealth generated by the portfolio sequence :math:`\mathbf{b}_{1:n}` at the end of the :math:`n`-th trading periods.

        .. math::

            \mathrm{FCW}=S_0 \prod_{t=1}^n \mathbf{b}_t^{\\top} \mathbf{x}_t=S_0 \prod_{t=1}^n \sum_{i=1}^m b_{t, i} x_{t, i},

        where :math:`S_0` is the initial cumulative wealth, :math:`\mathbf{b}_t` is the portfolio vector and :math:`\mathbf{x}_t` is the price relative vector in period :math:`t`.

        :param portfolios: Portfolio sequence :math:`\mathbf{b}_{1:n} \in \mathbb{R}^{n {\\times} m}`.
        :param labels: Price relative sequence :math:`\mathbf{x}_{1:n} \in \mathbb{R}_{+}^{n {\\times} m}`.
        :return: Tuple containing the daily return, daily cumulative wealth, and final cumulative wealth.
        """
        # Calculate daily returns
        daily_returns = np.sum(portfolios * labels, axis=1)
        # Calculate daily cumulative wealth
        daily_cumulative_wealth = np.cumprod(daily_returns)
        # Calculate final cumulative wealth
        final_cumulative_wealth = daily_cumulative_wealth[-1]

        return daily_returns, daily_cumulative_wealth, final_cumulative_wealth

    def caculate_portfolios(self) -> Tuple[np.ndarray, np.ndarray, float]:
        """
        Calculate the portfolio sequence :math:`\mathbf{b}_{1:n}`.

        :return: Tuple containing the portfolio sequence :math:`\mathbf{b}_{1:n}`, price relative sequence :math:`\mathbf{x}_{1:n}`, and other information.
        """
        model = torch.load(self.logdir + "/" + add_prefix("best_model.pt")).to(self.config["DEVICE"])
        model.eval()
        test_loader = self.load_dataset_output["test_loader"]
        num_trading_periods = self.load_dataset_output["num_test_periods"]  # Total number of trading periods
        num_assets = self.load_dataset_output["num_assets"]

        portfolios = torch.zeros((num_trading_periods, num_assets))
        labels = torch.zeros((num_trading_periods, num_assets))
        start_time = time.time()
        for i, data in enumerate(test_loader, 0):
            x_data, label = data
            labels[i, :] = label

            final_scores = model(x_data.float())
            portfolio = actual_portfolio_selection(final_scores)
            portfolios[i, :] = portfolio

        runtime = time.time() - start_time
        portfolios = portfolios.cpu().detach().numpy()
        labels = labels.cpu().detach().numpy()

        return portfolios, labels, runtime, num_trading_periods

    def caculate_runtime(self, runtime: float) -> float:
        """
        *Running Time* (RT): Measure the amount of time required to execute trades for an OLPS method. In live markets, any delays in processing or slow computation times can result in missed opportunities or executions at less favorable prices.

        .. note::

            While it is typically assumed in OLPS problems that execution at the desired price is always possible.

        :param runtime: *Running Time*.
        :return: *Running Time*.
        """
        return runtime

    def caculate_metric(self, portfolios=None, labels=None, runtime=None) -> Dict:
        """
        Calculate various performance metrics based on the provided portfolios, labels.

        :param portfolios: Portfolio sequence :math:`\mathbf{b}_{1:n}`.
        :param labels: Price relative sequence :math:`\mathbf{x}_{1:n}`.
        :param runtime: The runtime information. Required only when ``self.mode`` is ``ed`` (economic distillation).
        :return: Dictionary containing calculated performance metrics.
        """
        if self.mode == "normal":
            portfolios, labels, runtime, num_trading_periods = self.caculate_portfolios()
        elif self.mode == "ed":
            portfolios, labels = portfolios.cpu().detach().numpy(), labels.cpu().detach().numpy()
            num_trading_periods, num_assets = portfolios.shape

        # Profit Metrics
        daily_returns, daily_cumulative_wealth, final_cumulative_wealth = self.caculate_final_cumulative_wealth(portfolios, labels)
        annual_percentage_yield = self.caculate_annual_percentage_yield(num_trading_periods, final_cumulative_wealth)
        sharpe_ratio = self.caculate_sharpe_ratio(daily_returns, annual_percentage_yield)

        # Risk Metrics
        volatility_risk = self.caculate_volatility_risk(daily_returns)
        daily_drawdown, daily_maximum_drawdown, maximum_drawdown = self.caculate_maximum_drawdown(daily_cumulative_wealth)

        # Practical Metrics
        daily_turnover, transaction_costs_wealth = self.caculate_transaction_costs_wealth(portfolios, labels)
        average_turnover = self.caculate_average_turnover(daily_turnover)
        runtime = self.caculate_runtime(runtime)

        caculate_metric_output = {
            "logdir": self.logdir,
            "DCW": daily_cumulative_wealth,
            "CW": final_cumulative_wealth,
            "APY": annual_percentage_yield,
            "SR": sharpe_ratio,
            "VR": volatility_risk,
            "DDD": daily_drawdown,
            "DMDD": daily_maximum_drawdown,
            "MDD": maximum_drawdown,
            "ATO": average_turnover,
            "TCW": transaction_costs_wealth,
            "RT": runtime,
        }
        return caculate_metric_output


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.patches import Circle, RegularPolygon
from matplotlib.path import Path
from matplotlib.projections import register_projection
from matplotlib.projections.polar import PolarAxes
from matplotlib.spines import Spine
from matplotlib.transforms import Affine2D
from finol.utils import load_config, add_prefix


def radar_factory(num_vars, frame="circle"):
    """
    Create a radar chart with `num_vars` axes.

    This function creates a RadarAxes projection and registers it.

    Parameters
    ----------
    num_vars : int
        Number of variables for radar chart.
    frame : {"circle", "polygon"}
        Shape of frame surrounding axes.

    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2 * np.pi, num_vars, endpoint=False)

    class RadarTransform(PolarAxes.PolarTransform):

        def transform_path_non_affine(self, path):
            # Paths with non-unit interpolation steps correspond to gridlines,
            # in which case we force interpolation (to defeat PolarTransform's
            # autoconversion to circular arcs).
            if path._interpolation_steps > 1:
                path = path.interpolated(num_vars)
            return Path(self.transform(path.vertices), path.codes)

    class RadarAxes(PolarAxes):

        name = "radar"
        PolarTransform = RadarTransform

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # rotate plot such that the first axis is at the top
            self.set_theta_zero_location("N")

        def fill(self, *args, closed=True, **kwargs):
            """Override fill so that line is closed by default"""
            return super().fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super().plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            if x[0] != x[-1]:
                x = np.append(x, x[0])
                y = np.append(y, y[0])
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            # The Axes patch must be centered at (0.5, 0.5) and of radius 0.5
            # in axes coordinates.
            if frame == "circle":
                return Circle((0.5, 0.5), 0.5)
            elif frame == "polygon":
                return RegularPolygon((0.5, 0.5), num_vars,
                                      radius=.5, edgecolor="k")
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

        def _gen_axes_spines(self):
            if frame == "circle":
                return super()._gen_axes_spines()
            elif frame == "polygon":
                # spine_type must be "left"/"right"/"top"/"bottom"/"circle".
                spine = Spine(axes=self,
                              spine_type="circle",
                              path=Path.unit_regular_polygon(num_vars))
                # unit_regular_polygon gives a polygon of radius 1 centered at
                # (0, 0) but we want a polygon of radius 0.5 centered at (0.5,
                # 0.5) in axes coordinates.
                spine.set_transform(Affine2D().scale(.5).translate(.5, .5)
                                    + self.transAxes)
                return {"polar": spine}
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

    register_projection(RadarAxes)
    return theta


class ResultVisualizer:
    def __init__(self, load_benchmark_output) -> None:
        self.config = load_config()
        self.load_benchmark_output = load_benchmark_output
        self.top_5_baselines = load_benchmark_output["top_5_baselines"]
        self.logdir = load_benchmark_output["logdir"]

    def visualize_result(self) -> None:
        """
        Visualize the result using appropriate plotting methods.

        :return: None
        """
        # ["DCW", "DDD", "DMDD", "TCW"]
        for plot_type in ["DCW", "DMDD", "TCW", "RADAR"]:
            self.plot_type = plot_type
            if plot_type == "DCW":
                self.plot_daily_cumulative_wealth()
            elif plot_type == "DMDD":
                self.plot_daily_maximum_drawdown()
            elif plot_type == "TCW":
                self.plot_transaction_costs_adjusted_wealth()
            elif plot_type == "RADAR":
                self.plot_radar_chart()

    def plot_daily_cumulative_wealth(self) -> None:
        """
        Visualize the daily cumulative wealth comparison between the model and the top baselines.

        :return: None
        """
        df = self.load_benchmark_output["daily_cumulative_wealth"]
        df = df[self.top_5_baselines]
        num_columns = len(df.columns)

        plot_labels = {
            "en": ("Trading Periods", "Daily Cumulative Wealth"),
            "zh-CHS": ("交易期", "逐期累积财富"),
            "zh-CHT": ("交易期", "逐期累積財富")
        }
        xlabel, ylabel = plot_labels[self.config["PLOT_LANGUAGE"]]

        lines = ["-"] * (num_columns)
        colors = ["black"] * (num_columns - 1) + ["red"] * 1

        for i, column in enumerate(df.columns):
            plt.plot(
                df[column],
                linestyle=lines[i],
                color=colors[i],
                marker=self.config["MARKERS"][i],
                markevery=self.config["MARKEVERY"][self.config["DATASET_NAME"]],
                alpha=0.5,
                label=column
            )

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(self.config["DATASET_NAME"])
        plt.legend(loc="best")  # lower left upper left
        plt.grid(True)
        # plt.yscale("log")
        plt.tight_layout()
        plt.savefig(self.logdir + "/" + add_prefix(self.plot_type) + ".pdf", format="pdf", dpi=300, bbox_inches="tight")
        plt.show()
        plt.figure()

    def plot_daily_drawdown(self) -> None:
        """
        Visualize the daily drawdown comparison between the model and the top baselines.

        :return: None
        """
        df = self.load_benchmark_output["daily_daily_drawdown"]
        df = df[self.top_5_baselines]
        num_columns = len(df.columns)

        plot_labels = {
            "en": ("Trading Periods", "Daily DrawDown"),
            "zh-CHS": ("交易期", "逐期下行风险"),
            "zh-CHT": ("交易期", "逐期下行風險"),
        }
        xlabel, ylabel = plot_labels[self.config["PLOT_LANGUAGE"]]

        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"]:
            lines = ["-"] * (num_columns - 1) + [":"]
            colors = ["black"] * (num_columns - 2) + ["red"] * 2
        else:
            lines = ["-"] * (num_columns)
            colors = ["black"] * (num_columns - 1) + ["red"] * 1

        for i, column in enumerate(df.columns):
            plt.plot(
                df[column],
                linestyle=lines[i],
                color=colors[i],
                marker=self.config["MARKERS"][i],
                markevery=self.config["MARKEVERY"][self.config["DATASET_NAME"]],
                alpha=0.5,
                label=column
            )

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(self.config["DATASET_NAME"])
        plt.legend(loc="best")  # lower left upper left
        plt.grid(True)
        # plt.yscale("log")
        plt.tight_layout()
        plt.savefig(self.logdir + "/" + add_prefix(self.plot_type) + ".pdf", format="pdf", dpi=300, bbox_inches="tight")
        plt.show()
        plt.figure()

    def plot_daily_maximum_drawdown(self) -> None:
        """
        Visualize the daily maximum drawdown comparison between the model and the top baselines.

        :return: None
        """
        df = self.load_benchmark_output["daily_maximum_drawdown"]
        df = df[self.top_5_baselines]
        num_columns = len(df.columns)

        plot_labels = {
            "en": ("Trading Periods", "Daily Maximum DrawDown"),
            "zh-CHS": ("交易期", "逐期最大下行风险"),
            "zh-CHT": ("交易期", "逐期最大下行風險"),
        }
        xlabel, ylabel = plot_labels[self.config["PLOT_LANGUAGE"]]

        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"]:
            lines = ["-"] * (num_columns - 1) + [":"]
            colors = ["black"] * (num_columns - 2) + ["red"] * 2
        else:
            lines = ["-"] * (num_columns)
            colors = ["black"] * (num_columns - 1) + ["red"] * 1

        for i, column in enumerate(df.columns):
            plt.plot(
                df[column],
                linestyle=lines[i],
                color=colors[i],
                marker=self.config["MARKERS"][i],
                markevery=self.config["MARKEVERY"][self.config["DATASET_NAME"]],
                alpha=0.5,
                label=column
            )

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(self.config["DATASET_NAME"])
        plt.legend(loc="best")  # lower left upper left
        plt.grid(True)
        # plt.yscale("log")
        plt.tight_layout()
        plt.savefig(self.logdir + "/" + add_prefix(self.plot_type) + ".pdf", format="pdf", dpi=300, bbox_inches="tight")
        plt.show()
        plt.figure()

    def plot_transaction_costs_adjusted_wealth(self) -> None:
        """
        Visualize the transaction costs adjusted wealth comparison between the model and the top baselines.

        :return: None
        """
        df = self.load_benchmark_output["transaction_costs_adjusted_cumulative_wealth"]
        df = df[self.top_5_baselines]
        num_columns = len(df.columns)

        plot_labels = {
            "en": ("Transaction Costs Rates (%)", "Costs-Adjusted Cumulative Wealth"),
            "zh-CHS": ("交易费用率( %)", "考虑交易费用的累积财富"),
            "zh-CHT": ("交易費用率( %)", "考慮交易費用的累積財富"),
        }
        xlabel, ylabel = plot_labels[self.config["PLOT_LANGUAGE"]]

        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"]:
            lines = ["-"] * (num_columns - 1) + [":"]
            colors = ["black"] * (num_columns - 2) + ["red"] * 2
        else:
            lines = ["-"] * (num_columns)
            colors = ["black"] * (num_columns - 1) + ["red"] * 1

        # plt.xticks(ticks=[0, 2, 4, 6, 8, 10], labels=[0, 0.2, 0.4, 0.6, 0.8, 1])
        # plt.xticks(ticks=[0, 2.5, 5, 7.5, 10], labels=[0, 0.25, 0.5, 0.75, 1])
        plt.xticks(ticks=[0, 1.25, 2.5, 3.75, 5, 7.5, 10], labels=[0, 0.125, 0.25, 0.375, 0.5, 0.75, 1])

        # tc = 0.005
        # interval = 0.001
        for i, column in enumerate(df.columns):
            plt.plot(
                df[column].head(int(0.005/0.001)+1),
                linestyle=lines[i],
                color=colors[i],
                marker=self.config["MARKERS"][i],
                markevery=1,
                alpha=0.5,
                label=column
            )

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(self.config["DATASET_NAME"])
        plt.legend(loc="best")  # lower left upper left
        plt.grid(True)
        # plt.yscale("log")
        plt.tight_layout()
        plt.savefig(self.logdir + "/" + add_prefix(self.plot_type) + ".pdf", format="pdf", dpi=300, bbox_inches="tight")
        plt.show()
        plt.figure()

    def plot_radar_chart(self) -> None:
        """
        Visualize the radar chart comparison between the model and the top baselines.

        :return: None
        """
        final_profit_result = self.load_benchmark_output["final_profit_result"][["Metric"] + self.top_5_baselines].set_index("Metric")
        final_risk_result = self.load_benchmark_output["final_risk_result"][["Metric"] + self.top_5_baselines].set_index("Metric")

        df = pd.concat([final_profit_result, final_risk_result], ignore_index=False)
        num_columns = len(df.columns)

        data = []
        for strategy in self.top_5_baselines:
            _ = [df[strategy]["CW"], -df[strategy]["MDD"], -df[strategy]["VR"], df[strategy]["SR"], df[strategy]["APY"]]
            data += [_]

        # scale the data to [0, 1]
        data_array = np.array(data)
        min_val = np.min(data_array, axis=0)
        max_val = np.max(data_array, axis=0)
        scaled_data = (data_array - min_val) / (max_val - min_val)

        theta = radar_factory(num_vars=5, frame="polygon")
        labels = ["CW", "- MDD", "- VR", "SR", "APY"]

        fig, ax = plt.subplots(subplot_kw=dict(projection="radar"))
        colors = ["black"] * (num_columns - 1) + ["red"] * 1

        # plot the four cases from the example data on separate axes
        ax.set_rgrids([0.2, 0.4, 0.6, 0.8])

        for d, color, marker in zip(scaled_data, colors, self.config["MARKERS"]):
            ax.plot(theta, d, color=color, alpha=0.5, marker=marker)
            ax.fill(theta, d, facecolor=color, alpha=0.15, label="_nolegend_")
        ax.set_varlabels(labels)

        # add legend relative to top-left plot
        plt.title(self.config["DATASET_NAME"])
        ax.legend(df.columns, loc=(0.7, 0.75))
        plt.tight_layout()
        plt.savefig(self.logdir + "/" + add_prefix(self.plot_type) + ".pdf", format="pdf", dpi=300, bbox_inches="tight")
        plt.show()



import matplotlib.pyplot as plt

from rich import print
from finol.evaluation_layer.benchmark_loader import BenchmarkLoader
from finol.evaluation_layer.distiller_selector import DistillerSelector
from finol.evaluation_layer.economic_distiller import EconomicDistiller
from finol.evaluation_layer.metric_caculator import MetricCaculator
from finol.evaluation_layer.model_evaluator import ModelEvaluator
from finol.evaluation_layer.result_visualizer import ResultVisualizer
from finol.utils import load_config

config = load_config()
if config["PLOT_LANGUAGE"].startswith('zh'):
    plt.rcParams["font.family"] = "Microsoft YaHei"

__all__ = [
    "BenchmarkLoader",
    "DistillerSelector",
    "EconomicDistiller",
    "MetricCaculator",
    "ModelEvaluator",
    "ResultVisualizer",
]



import torch

from typing import List, Tuple, Dict, Union
from sklearn import preprocessing
from finol.utils import load_config


class ScalerSelector:
    """
    Class to select different types of scalers and performs partial data standardization depending on the configuration.
    """
    def __init__(self) -> None:
        self.config = load_config()
        self.scaler_dict = {
            "None": "None",
            "StandardScaler": preprocessing.StandardScaler,
            "MinMaxScaler": preprocessing.MinMaxScaler,
            "MaxAbsScaler": preprocessing.MaxAbsScaler,
            "RobustScaler": preprocessing.RobustScaler,
            "WindowStandardScaler": self.window_normalize_via_StandardScaler,
            "WindowMinMaxScaler": self.window_normalize_via_MinMaxScaler,
            "WindowMaxAbsScaler": self.window_normalize_via_MaxAbsScaler,
            "WindowRobustScaler": self.window_normalize_via_RobustScaler,
        }

    def select_scaler(self) -> Union[object, None]:
        """
        Select the data scaler based on the configuration.

        :return: The selected scaler object, or None if no scaler is specified.
        """
        scaler_cls = self.scaler_dict.get(self.config["SCALER"], None)
        if scaler_cls is None:
            raise ValueError(f"Invalid scaler name: {self.config['SCALER']}. Supported scalers are: {self.scaler_dict.keys()}")

        if self.config["SCALER"] == "None" or self.config["SCALER"].startswith("Window"):
            return None

        scaler = scaler_cls()
        return scaler

    def window_normalize_via_StandardScaler(self, data: torch.Tensor) -> torch.Tensor:
        """
        Normalize the input tensor data along the window dimension via StandardScaler.

        :param data: Input data tensor of shape (..., window_size, feature_dim).
        :return: Standardized data tensor of the same shape.
        """
        window_mean = torch.mean(data, dim=-2, keepdim=True)
        window_std = torch.std(data, dim=-2, keepdim=True) + 1e-7
        standardized_data = (data - window_mean) / window_std
        return standardized_data

    def window_normalize_via_MinMaxScaler(self, data: torch.Tensor) -> torch.Tensor:
        """
        Normalize the input tensor data along the window dimension via MinMaxScaler.

        :param data: Input data tensor of shape (..., window_size, feature_dim).
        :return: Standardized data tensor of the same shape.
        """
        min_vals = torch.min(data, dim=-2, keepdim=True).values
        max_vals = torch.max(data, dim=-2, keepdim=True).values
        normalized_data = (data - min_vals) / (max_vals - min_vals + 1e-7)
        return normalized_data

    def window_normalize_via_MaxAbsScaler(self, data: torch.Tensor) -> torch.Tensor:
        """
        Normalize the input tensor data along the window dimension via MaxAbsScaler.

        :param data: Input data tensor of shape (..., window_size, feature_dim).
        :return: Standardized data tensor of the same shape.
        """
        max_abs_vals = torch.max(torch.abs(data), dim=-2, keepdim=True).values
        normalized_data = data / (max_abs_vals + 1e-7)
        return normalized_data

    def window_normalize_via_RobustScaler(self, data: torch.Tensor) -> torch.Tensor:
        """
        Normalize the input tensor data along the window dimension via RobustScaler.

        :param data: Input data tensor of shape (..., window_size, feature_dim).
        :return: Standardized data tensor of the same shape.
        """

        # Calculate the median and IQR (interquartile range) along the window dimension
        window_median = torch.median(data, dim=-2, keepdim=True).values
        q1 = torch.quantile(data, 0.25, dim=-2, keepdim=True)
        q3 = torch.quantile(data, 0.75, dim=-2, keepdim=True)
        window_iqr = q3 - q1 + 1e-7

        # Scale the data using the median and IQR
        normalized_data = (data - window_median) / window_iqr
        return normalized_data

    def window_normalize(self, data: torch.Tensor) -> torch.Tensor:
        """
        Normalize the input tensor data along the window dimension based on different methods.

        :param data: Input data tensor of shape (..., window_size, feature_dim).
        :return: Standardized data tensor of the same shape.
        """
        scaler_cls = self.scaler_dict.get(self.config["SCALER"], None)
        if scaler_cls is None:
            raise ValueError(f"Invalid scaler name: {self.config['SCALER']}. Supported scalers are: {self.scaler_dict.keys()}")

        normalized_data = scaler_cls(data)
        return normalized_data





__version__ = "0.1.28"
__author__ = "FinOL Contributors"

#
"""
python setup.py sdist build
twine upload dist/*

pip install --upgrade finol
python -m finol.FinOLAPP
python -m finol.main.main.py
"""




import os
import time
import torch
import pandas as pd

from tqdm import tqdm
from typing import List, Tuple, Dict, Union
from torch.utils.data import TensorDataset
from torch.utils.data import DataLoader
from finol.data_layer.scaler_selector import ScalerSelector
from finol.utils import ROOT_PATH, load_config, update_config, make_logdir, check_update, download_data


class DatasetLoader:
    """
    Class to load different types of datasets.

    """
    def __init__(self) -> None:
        self.config = load_config()
        check_update()
        download_data()

    def access_data(self, folder_path: str) -> List[pd.DataFrame]:
        """
        Load raw data files from a specified folder path and return a list of DataFrames.

        :param folder_path: Path to the folder containing raw data files.
        :return: List of DataFrames containing the loaded raw data.
        """
        raw_files = []
        for file_name in tqdm(os.listdir(folder_path), desc="Data Loading"):
            if file_name.endswith(".xlsx") or file_name.endswith(".xls"):
                file_path = os.path.join(folder_path, file_name)
                try:
                    dataframe = pd.read_excel(file_path)
                    raw_files.append(dataframe)
                except Exception as e:
                    print(f"An error occurred while loading file {file_path}: {str(e)}")
        return raw_files

    def feature_engineering(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, List[str], Dict[str, int]]:
        """
        Perform feature engineering on the input DataFrame to generate various types of features.

        :param df: Input DataFrame to be engineered.
        :return: Tuple containing the engineered DataFrame, detailed feature list, and number of features in each category.
        """
        ohlcv_features_df = pd.DataFrame()
        overlap_features_df = pd.DataFrame()
        momentum_features_df = pd.DataFrame()
        volume_features_df = pd.DataFrame()
        cycle_features_df = pd.DataFrame()
        price_features_df = pd.DataFrame()
        volatility_features_df = pd.DataFrame()
        pattern_features_df = pd.DataFrame()

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OHLCV_FEATURES"]:
            ohlcv_features = {
                "OPEN": df.OPEN,
                "HIGH": df.HIGH,
                "LOW": df.LOW,
                "CLOSE": df.CLOSE,
                "VOLUME": df.VOLUME
            }
            ohlcv_features_df = pd.DataFrame(ohlcv_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OVERLAP_FEATURES"]:
            import talib as ta
            overlap_features = {
                "BBANDS_UPPER": ta.BBANDS(df.CLOSE)[0],  # Bollinger Bands - Upper Band
                "BBANDS_MIDDLE": ta.BBANDS(df.CLOSE)[1],  # Bollinger Bands - Middle Band
                "BBANDS_LOWER": ta.BBANDS(df.CLOSE)[2],  # Bollinger Bands - Lower Band
                "DEMA": ta.DEMA(df.CLOSE),  # Double Exponential Moving Average
                "EMA": ta.EMA(df.CLOSE),  # Exponential Moving Average
                "HT_TRENDLINE": ta.HT_TRENDLINE(df.CLOSE),  # Hilbert Transform - Instantaneous Trendline
                "KAMA": ta.KAMA(df.CLOSE),  # Kaufman Adaptive Moving Average
                "MA": ta.MA(df.CLOSE),  # Moving Average
                "MAMA": ta.MAMA(df.CLOSE)[0],  # MESA Adaptive Moving Average - MAMA
                "MAMA_FAMA": ta.MAMA(df.CLOSE)[1],  # MESA Adaptive Moving Average - FAMA
                "MAVP": ta.MAVP(df.CLOSE, df.DATE),  # Moving Average with Variable Period
                "MIDPOINT": ta.MIDPOINT(df.CLOSE),  # MidPoint over Period
                "MIDPRICE": ta.MIDPRICE(df.HIGH, df.LOW),  # Midpoint Price over Period
                "SAR": ta.SAR(df.HIGH, df.LOW),  # Parabolic SAR
                "SAREXT": ta.SAREXT(df.HIGH, df.LOW),  # Parabolic SAR - Extended
                "SMA": ta.SMA(df.CLOSE),  # Simple Moving Average
                "T3": ta.T3(df.CLOSE),  # Triple Exponential Moving Average (T3)
                "TEMA": ta.TEMA(df.CLOSE),  # Triple Exponential Moving Average
                "TRIMA": ta.TRIMA(df.CLOSE),  # Triangular Moving Average
                "WMA": ta.WMA(df.CLOSE)  # Weighted Moving Average
            }
            overlap_features_df = pd.DataFrame(overlap_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_MOMENTUM_FEATURES"]:
            import talib as ta
            momentum_features = {
                "ADX": ta.ADX(df.HIGH, df.LOW, df.CLOSE),  # Average Directional Movement Index
                "ADXR": ta.ADXR(df.HIGH, df.LOW, df.CLOSE),  # Average Directional Movement Index Rating
                "APO": ta.APO(df.CLOSE),  # Absolute Price Oscillator
                "AROON_UP": ta.AROON(df.HIGH, df.LOW)[0],  # Aroon Up
                "AROON_DOWN": ta.AROON(df.HIGH, df.LOW)[1],  # Aroon Down
                "AROONOSC": ta.AROONOSC(df.HIGH, df.LOW),  # Aroon Oscillator
                "BOP": ta.BOP(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Balance Of Power
                "CCI": ta.CCI(df.HIGH, df.LOW, df.CLOSE),  # Commodity Channel Index
                "CMO": ta.CMO(df.CLOSE),  # Chande Momentum Oscillator
                "DX": ta.DX(df.HIGH, df.LOW, df.CLOSE),  # Directional Movement Index
                "MACD": ta.MACD(df.CLOSE)[0],  # Moving Average Convergence/Divergence
                "MACD_SIGNAL": ta.MACD(df.CLOSE)[1],  # MACD Signal Line
                "MACD_HIST": ta.MACD(df.CLOSE)[2],  # MACD Histogram
                "MACDEXT": ta.MACDEXT(df.CLOSE)[0],  # MACD with controllable MA type
                "MACDEXT_SIGNAL": ta.MACDEXT(df.CLOSE)[1],  # MACDEXT Signal Line
                "MACDEXT_HIST": ta.MACDEXT(df.CLOSE)[2],  # MACDEXT Histogram
                "MACDFIX": ta.MACDFIX(df.CLOSE)[0],  # Moving Average Convergence/Divergence Fix 12/26
                "MACDFIX_SIGNAL": ta.MACDFIX(df.CLOSE)[1],  # MACDFIX Signal Line
                "MACDFIX_HIST": ta.MACDFIX(df.CLOSE)[2],  # MACDFIX Histogram
                "MFI": ta.MFI(df.HIGH, df.LOW, df.CLOSE, df.VOLUME),  # Money Flow Index
                "MINUS_DI": ta.MINUS_DI(df.HIGH, df.LOW, df.CLOSE),  # Minus Directional Indicator
                "MINUS_DM": ta.MINUS_DM(df.HIGH, df.LOW),  # Minus Directional Movement
                "MOM": ta.MOM(df.CLOSE),  # Momentum
                "PLUS_DI": ta.PLUS_DI(df.HIGH, df.LOW, df.CLOSE),  # Plus Directional Indicator
                "PLUS_DM": ta.PLUS_DM(df.HIGH, df.LOW),  # Plus Directional Movement
                "PPO": ta.PPO(df.CLOSE),  # Percentage Price Oscillator
                "ROC": ta.ROC(df.CLOSE),  # Rate of change: ((price/prevPrice)-1)*100
                "ROCP": ta.ROCP(df.CLOSE),  # Rate of change Percentage: (price-prevPrice)/prevPrice
                "ROCR": ta.ROCR(df.CLOSE),  # Rate of change ratio: (price/prevPrice)
                "ROCR100": ta.ROCR100(df.CLOSE),  # Rate of change ratio 100 scale: (price/prevPrice)*100
                "RSI": ta.RSI(df.CLOSE),  # Relative Strength Index
                "STOCH_K": ta.STOCH(df.HIGH, df.LOW, df.CLOSE)[0],  # Stochastic %K
                "STOCH_D": ta.STOCH(df.HIGH, df.LOW, df.CLOSE)[1],  # Stochastic %D
                "STOCHF_K": ta.STOCHF(df.HIGH, df.LOW, df.CLOSE)[0],  # Stochastic Fast %K
                "STOCHF_D": ta.STOCHF(df.HIGH, df.LOW, df.CLOSE)[1],  # Stochastic Fast %D
                "STOCHRSI_K": ta.STOCHRSI(df.CLOSE)[0],  # Stochastic RSI %K
                "STOCHRSI_D": ta.STOCHRSI(df.CLOSE)[1],  # Stochastic RSI %D
                "TRIX": ta.TRIX(df.CLOSE),  # 1-day Rate-Of-Change (ROC) of a Triple Smooth EMA
                "ULTOSC": ta.ULTOSC(df.HIGH, df.LOW, df.CLOSE),  # Ultimate Oscillator
                "WILLR": ta.WILLR(df.HIGH, df.LOW, df.CLOSE)  # Williams' %R
            }
            momentum_features_df = pd.DataFrame(momentum_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLUME_FEATURES"]:
            import talib as ta
            volume_features = {
                "AD": ta.AD(df.HIGH, df.LOW, df.CLOSE, df.VOLUME),  # Chaikin A/D Line
                "ADOSC": ta.ADOSC(df.HIGH, df.LOW, df.CLOSE, df.VOLUME),  # Chaikin A/D Oscillator
                "OBV": ta.OBV(df.CLOSE, df.VOLUME)  # On Balance Volume
            }
            volume_features_df = pd.DataFrame(volume_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_CYCLE_FEATURES"]:
            import talib as ta
            cycle_features = {
                "HT_DCPERIOD": ta.HT_DCPERIOD(df.CLOSE),  # Hilbert Transform - Dominant Cycle Period
                "HT_DCPHASE": ta.HT_DCPHASE(df.CLOSE),  # Hilbert Transform - Dominant Cycle Phase
                "HT_PHASOR_INPHASE": ta.HT_PHASOR(df.CLOSE)[0],  # Hilbert Transform - Phasor Components
                "HT_PHASOR_QUADRATURE": ta.HT_PHASOR(df.CLOSE)[1],  # Hilbert Transform - Phasor Components
                "HT_SINE_LEADSINE": ta.HT_SINE(df.CLOSE)[0],  # Hilbert Transform - SineWave
                "HT_SINE_SINEWAVE": ta.HT_SINE(df.CLOSE)[1],  # Hilbert Transform - SineWave
                "HT_TRENDMODE": ta.HT_TRENDMODE(df.CLOSE)  # Hilbert Transform - Trend vs Cycle Mode
            }
            cycle_features_df = pd.DataFrame(cycle_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PRICE_FEATURES"]:
            import talib as ta
            price_features = {
                "AVGPRICE": ta.AVGPRICE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Average Price
                "MEDPRICE": ta.MEDPRICE(df.HIGH, df.LOW),  # Median Price
                "TYPPRICE": ta.TYPPRICE(df.HIGH, df.LOW, df.CLOSE),  # Typical Price
                "WCLPRICE": ta.WCLPRICE(df.HIGH, df.LOW, df.CLOSE)  # Weighted Close Price
            }
            price_features_df = pd.DataFrame(price_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLATILITY_FEATURES"]:
            import talib as ta
            volatility_features = {
                "ATR": ta.ATR(df.HIGH, df.LOW, df.CLOSE),  # Average True Range
                "NATR": ta.NATR(df.HIGH, df.LOW, df.CLOSE),  # Normalized Average True Range
                "TRANGE": ta.TRANGE(df.HIGH, df.LOW, df.CLOSE)  # True Range
            }
            volatility_features_df = pd.DataFrame(volatility_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PATTERN_FEATURES"]:
            import talib as ta
            pattern_features = {
                "CDL2CROWS": ta.CDL2CROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Two Crows
                "CDL3BLACKCROWS": ta.CDL3BLACKCROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Black Crows
                "CDL3INSIDE": ta.CDL3INSIDE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Inside Up/Down
                "CDL3LINESTRIKE": ta.CDL3LINESTRIKE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three-Line Strike
                "CDL3OUTSIDE": ta.CDL3OUTSIDE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Outside Up/Down
                "CDL3STARSINSOUTH": ta.CDL3STARSINSOUTH(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Stars In The South
                "CDL3WHITESOLDIERS": ta.CDL3WHITESOLDIERS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Advancing White Soldiers
                "CDLABANDONEDBABY": ta.CDLABANDONEDBABY(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Abandoned Baby
                "CDLADVANCEBLOCK": ta.CDLADVANCEBLOCK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Advance Block
                "CDLBELTHOLD": ta.CDLBELTHOLD(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Belt-Hold
                "CDLBREAKAWAY": ta.CDLBREAKAWAY(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Breakaway
                "CDLCLOSINGMARUBOZU": ta.CDLCLOSINGMARUBOZU(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Closing Marubozu
                "CDLCONCEALBABYSWALL": ta.CDLCONCEALBABYSWALL(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Concealing Baby Swallow
                "CDLCOUNTERATTACK": ta.CDLCOUNTERATTACK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Counterattack
                "CDLDARKCLOUDCOVER": ta.CDLDARKCLOUDCOVER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Dark Cloud Cover
                "CDLDOJI": ta.CDLDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Doji,
                "CDLDOJISTAR": ta.CDLDOJISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Doji Star
                "CDLDRAGONFLYDOJI": ta.CDLDRAGONFLYDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Dragonfly Doji
                "CDLENGULFING": ta.CDLENGULFING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Engulfing Pattern
                "CDLEVENINGDOJISTAR": ta.CDLEVENINGDOJISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Evening Doji Star
                "CDLEVENINGSTAR": ta.CDLEVENINGSTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Evening Star
                "CDLGAPSIDESIDEWHITE": ta.CDLGAPSIDESIDEWHITE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Up/Down-Gap Side-By-Side White Lines
                "CDLGRAVESTONEDOJI": ta.CDLGRAVESTONEDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Gravestone Doji
                "CDLHAMMER": ta.CDLHAMMER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Hammer
                "CDLHANGINGMAN": ta.CDLHANGINGMAN(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Hanging Man
                "CDLHARAMI": ta.CDLHARAMI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Harami Pattern
                "CDLHARAMICROSS": ta.CDLHARAMICROSS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Harami Cross Pattern
                "CDLHIGHWAVE": ta.CDLHIGHWAVE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # High-Wave Candle
                "CDLHIKKAKE": ta.CDLHIKKAKE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Hikkake Pattern
                "CDLHIKKAKEMOD": ta.CDLHIKKAKEMOD(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Modified Hikkake Pattern
                "CDLHOMINGPIGEON": ta.CDLHOMINGPIGEON(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Homing Pigeon
                "CDLIDENTICAL3CROWS": ta.CDLIDENTICAL3CROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Identical Three Crows
                "CDLINNECK": ta.CDLINNECK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # In-Neck Pattern
                "CDLINVERTEDHAMMER": ta.CDLINVERTEDHAMMER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Inverted Hammer
                "CDLKICKING": ta.CDLKICKING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Kicking
                "CDLKICKINGBYLENGTH": ta.CDLKICKINGBYLENGTH(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Kicking - Bull/Bear Determined by the Longer Marubozu
                "CDLLADDERBOTTOM": ta.CDLLADDERBOTTOM(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Ladder Bottom
                "CDLLONGLEGGEDDOJI": ta.CDLLONGLEGGEDDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Long Legged Doji
                "CDLLONGLINE": ta.CDLLONGLINE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Long Line Candle
                "CDLMARUBOZU": ta.CDLMARUBOZU(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Marubozu
                "CDLMATCHINGLOW": ta.CDLMATCHINGLOW(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Matching Low
                "CDLMATHOLD": ta.CDLMATHOLD(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Mat Hold
                "CDLMORNINGDOJISTAR": ta.CDLMORNINGDOJISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Morning Doji Star
                "CDLMORNINGSTAR": ta.CDLMORNINGSTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Morning Star
                "CDLONNECK": ta.CDLONNECK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # On-Neck Pattern
                "CDLPIERCING": ta.CDLPIERCING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Piercing Pattern
                "CDLRICKSHAWMAN": ta.CDLRICKSHAWMAN(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Rickshaw Man
                "CDLRISEFALL3METHODS": ta.CDLRISEFALL3METHODS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Rising/Falling Three Methods
                "CDLSEPARATINGLINES": ta.CDLSEPARATINGLINES(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Separating Lines
                "CDLSHOOTINGSTAR": ta.CDLSHOOTINGSTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Shooting Star
                "CDLSHORTLINE": ta.CDLSHORTLINE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Short Line Candle
                "CDLSPINNINGTOP": ta.CDLSPINNINGTOP(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Spinning Top
                "CDLSTALLEDPATTERN": ta.CDLSTALLEDPATTERN(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Stalled Pattern
                "CDLSTICKSANDWICH": ta.CDLSTICKSANDWICH(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Stick Sandwich
                "CDLTAKURI": ta.CDLTAKURI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Takuri (Dragonfly Doji with Very Long Lower Shadow)
                "CDLTASUKIGAP": ta.CDLTASUKIGAP(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Tasuki Gap
                "CDLTHRUSTING": ta.CDLTHRUSTING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Thrusting Pattern
                "CDLTRISTAR": ta.CDLTRISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Tristar Pattern
                "CDLUNIQUE3RIVER": ta.CDLUNIQUE3RIVER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Unique 3 River
                "CDLUPSIDEGAP2CROWS": ta.CDLUPSIDEGAP2CROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Upside Gap Two Crows
                "CDLXSIDEGAP3METHODS": ta.CDLXSIDEGAP3METHODS(df.OPEN, df.HIGH, df.LOW, df.CLOSE)  # Upside/Downside Gap Three Methods
            }
            pattern_features_df = pd.DataFrame(pattern_features)

        _ = pd.concat([df["DATE"], ohlcv_features_df, overlap_features_df, momentum_features_df, volume_features_df, cycle_features_df,
                        price_features_df, volatility_features_df, pattern_features_df], axis=1)
        _.set_index("DATE", inplace=True)
        DETAILED_FEATURE_LIST = _.columns.tolist()
        DETAILED_NUM_FEATURES = {
            "OHLCV_FEATURES": ohlcv_features_df.shape[1],
            "OVERLAP_FEATURES": overlap_features_df.shape[1],
            "MOMENTUM_FEATURES": momentum_features_df.shape[1],
            "VOLUME_FEATURES": volume_features_df.shape[1],
            "CYCLE_FEATURES": cycle_features_df.shape[1],
            "PRICE_FEATURES": price_features_df.shape[1],
            "VOLATILITY_FEATURES": volatility_features_df.shape[1],
            "PATTERN_FEATURES": pattern_features_df.shape[1],
        }
        return _, DETAILED_FEATURE_LIST, DETAILED_NUM_FEATURES

    def augment_data(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, int]:
        """
        Augment the provided DataFrame based on the configuration.

        :param df: Input DataFrame to be augmented.
        :return: Augmented DataFrame with window data and the window size used for augmentation.
        """
        if self.config["DATA_AUGMENTATION_CONFIG"]["WINDOW_DATA"]["INCLUDE_WINDOW_DATA"]:
            WINDOW_SIZE = self.config["DATA_AUGMENTATION_CONFIG"]["WINDOW_DATA"]["WINDOW_SIZE"]
            df = pd.concat([df] + [df.shift(i).add_prefix(f"prev_{i}_") for i in range(1, WINDOW_SIZE)], axis=1)
        else:
            WINDOW_SIZE = 1
        return df, WINDOW_SIZE

    def clean_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Clean the DataFrame by removing rows with missing values.

        :param df: Input DataFrame to be cleaned.
        :return: DataFrame with rows containing any missing values removed.
        """
        return df.dropna(how="any")

    def calculate_zscore(self, df: pd.DataFrame) -> Union[object, None]:
        """
        Calculate the z-scores for numeric features in the provided DataFrame.

        :param df: DataFrame containing the data for z-score calculation.
        :return: Z-score object for the numeric features in the DataFrame, in some cases the return can be None.
        """
        df_normalized = df.copy()
        numeric_features = df.select_dtypes(include=["int", "float"]).columns
        scaler = ScalerSelector().select_scaler()
        if scaler != None:
            zscore = scaler.fit(df_normalized[numeric_features])  # zscore is different for different assets
        else:
            zscore = None
        return zscore

    def normalize_data(self, df: pd.DataFrame, zscore: object) -> pd.DataFrame:
        """
        Normalize all numeric features in DataFrame.

        :param df: Input DataFrame to be normalized.
        :param zscore: Z-score object used for normalization.
        :return: DataFrame with normalized numeric features.
        """
        df_normalized = df.copy()
        numeric_features = df.select_dtypes(include=["int", "float"]).columns
        if zscore != None:
            df_normalized[numeric_features] = zscore.transform(df_normalized[numeric_features])
            # print(df_normalized[numeric_features])
            # print(df_normalized[numeric_features]["OPEN"].values)
            # print(df_normalized[numeric_features]["OPEN"].values.mean())  # -3.620599927440001e-16
            # print(df_normalized[numeric_features]["OPEN"].values.std())  # 1
        else:
            df_normalized[numeric_features] = df_normalized[numeric_features]
        return df_normalized

    def split_data(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Split the DataFrame into train, validation, and test sets.

        :param df: Input DataFrame to be split.
        :return: Tuple containing the train, validation, and test DataFrames.
        """
        train_start = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["TRAIN_START_TIMESTAMP"])
        train_end = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["TRAIN_END_TIMESTAMP"])
        val_start = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["VAL_START_TIMESTAMP"])
        val_end = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["VAL_END_TIMESTAMP"])
        test_start = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["TEST_START_TIMESTAMP"])
        test_end = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["TEST_END_TIMESTAMP"])

        train = df[df.index <= train_end]
        val = df[(df.index >= val_start) & (df.index <= val_end)]
        test = df[(df.index >= test_start) & (df.index <= test_end)]
        return train, val, test

    def make_label(self, raw_df: pd.DataFrame, df: pd.DataFrame) -> pd.DataFrame:
        """
        Generate labels, i.e. price relatives.

        :param raw_df: Raw DataFrame containing 'CLOSE' prices.
        :param df: DataFrame to merge the labels with.
        :return: DataFrame containing the generated labels.
        """
        raw_df.set_index("DATE", inplace=True)
        df_label = pd.DataFrame({"LABEL": raw_df.CLOSE.shift(-1) / raw_df.CLOSE})
        df_label.iloc[-1, -1] = 1

        df_label = pd.merge(df, df_label, left_index=True, right_index=True, how="outer")
        df_label = df_label.dropna(how="any")
        df_label = df_label["LABEL"].to_frame()
        return df_label

    def load_dataset(self) -> Dict:
        """
        Load the raw data, perform data pre-processing operations, and prepare DataLoader for training, validation, and testing.

        :return: Dictionary containing various data loaders and information about the dataset.
        """
        # print(self.config["MANUAL_SEED"])
        logdir = make_logdir()

        if self.config["LOAD_LOCAL_DATALOADER"]:
            try:
                load_dataset_output = torch.load(ROOT_PATH + "/data/datasets/" + self.config["DATASET_NAME"] + "_load_dataset_output.pt")
                load_dataset_output["logdir"] = logdir
                print("Local dataloader loaded successfully!")
            except Exception as e:
                self.config["LOAD_LOCAL_DATALOADER"] = False
                update_config(self.config)
                print(f"Local dataloader failed: {e}, the config['LOAD_LOCAL_DATALOADER'] is modified as `False` "
                      "automatically")

        # Ensure that the config is updated in all cases
        update_config(self.config)

        # Clear data structures if LOAD_LOCAL_DATALOADER is False
        if not self.config["LOAD_LOCAL_DATALOADER"]:
            ds_train = []
            ds_val = []
            ds_test = []
            label_train = []
            label_val = []
            label_test = []
            df_label_MATLAB = pd.DataFrame()

            raw_files = self.access_data(ROOT_PATH + "/data/datasets/" + self.config["DATASET_NAME"])
            if len(raw_files) > 0:
                # feature_engineering(excel_files)
                for i, df in tqdm(enumerate(raw_files), total=len(raw_files), desc="Data Processing"):
                    raw_df = df.copy()
                    df = self.clean_data(df)

                    df, DETAILED_FEATURE_LIST, DETAILED_NUM_FEATURES = self.feature_engineering(df)
                    df = self.clean_data(df)

                    df, WINDOW_SIZE = self.augment_data(df)
                    df = self.clean_data(df)

                    df_label = self.make_label(raw_df, df)

                    # df_label_temp = df_label.copy()
                    # test_start = pd.to_datetime(DATASET_SPLIT_CONFIG.get(dataset_name)["TEST_START_TIMESTAMP"])
                    # test_end = pd.to_datetime(DATASET_SPLIT_CONFIG.get(dataset_name)["TEST_END_TIMESTAMP"])
                    # df_label_temp = df_label_temp[(df_label_temp.index >= test_start) & (df_label_temp.index <= test_end)]
                    # df_label_MATLAB = pd.concat([df_label_MATLAB, df_label_temp], axis=1)

                    train, val, test = self.split_data(df)
                    train_label, val_label, test_label = self.split_data(df_label)

                    zscore_train = self.calculate_zscore(train)
                    train_normalization = self.normalize_data(train, zscore_train)
                    val_normalization = self.normalize_data(val, zscore_train)
                    test_normalization = self.normalize_data(test, zscore_train)

                    train_normalization = self.clean_data(train_normalization)
                    val_normalization = self.clean_data(val_normalization)
                    test_normalization = self.clean_data(test_normalization)

                    ds_train.append(torch.from_numpy(train_normalization.values))
                    ds_val.append(torch.from_numpy(val_normalization.values))
                    ds_test.append(torch.from_numpy(test_normalization.values))
                    label_train.append(torch.from_numpy(train_label["LABEL"].values))
                    label_val.append(torch.from_numpy(val_label["LABEL"].values))
                    label_test.append(torch.from_numpy(test_label["LABEL"].values))

                ds_train = torch.stack(ds_train).permute(1, 0, 2).to(self.config["DEVICE"])  # [num_assets, num_train_periods, num_feats] -> [num_train_periods, num_assets, num_feats]
                ds_val = torch.stack(ds_val).permute(1, 0, 2).to(self.config["DEVICE"])
                ds_test = torch.stack(ds_test).permute(1, 0, 2).to(self.config["DEVICE"])
                label_train = torch.stack(label_train).transpose(0, 1).to(self.config["DEVICE"])  # [num_assets, num_train_periods] -> [num_train_periods, num_assets]
                label_val = torch.stack(label_val).transpose(0, 1).to(self.config["DEVICE"])
                label_test = torch.stack(label_test).transpose(0, 1).to(self.config["DEVICE"])

                train_ids = TensorDataset(ds_train, label_train)
                val_ids = TensorDataset(ds_val, label_val)
                test_ids = TensorDataset(ds_test, label_test)

                # Save the price relative tensor to a mat file with the data set name
                # from scipy import io
                # io.savemat("price_relative_" + dataset_name + ".mat", {"data": df_label_MATLAB.values})
            else:
                print("No Excel files found.")

            load_dataset_output = {
                "logdir": logdir,
                "train_loader": DataLoader(train_ids, batch_size=self.config["BATCH_SIZE"][self.config["DATASET_NAME"]], shuffle=False, drop_last=False),
                "val_loader": DataLoader(val_ids, batch_size=self.config["BATCH_SIZE"][self.config["DATASET_NAME"]], shuffle=False, drop_last=False),
                "test_loader": DataLoader(test_ids, batch_size=1, shuffle=False, drop_last=False),
                "test_loader_for_train": DataLoader(test_ids, batch_size=self.config["BATCH_SIZE"][self.config["DATASET_NAME"]], shuffle=False, drop_last=False),
                "num_train_periods": ds_train.shape[0],
                "num_val_periods": ds_val.shape[0],
                "num_test_periods": ds_test.shape[0],
                "num_assets": ds_train.shape[1],
                "num_features_augmented": ds_train.shape[2],
                "num_features_original": int(ds_train.shape[2] / (WINDOW_SIZE)),
                "detailed_num_features": DETAILED_NUM_FEATURES,
                "window_size": WINDOW_SIZE,
                "overall_feature_list": [feature[8:] for feature, include in self.config["FEATURE_ENGINEERING_CONFIG"].items() if include],
                "detailed_feature_list": DETAILED_FEATURE_LIST,

            }
            torch.save(load_dataset_output, ROOT_PATH + "/data/datasets/" + self.config["DATASET_NAME"] + "_load_dataset_output.pt")
        # print(load_dataset_output)
        return load_dataset_output


if __name__ == "__main__":
    load_dataset_output = DatasetLoader().load_dataset()
    # print(load_dataset_output)


import time
import warnings
import pandas as pd
import numpy as np


from tabulate import tabulate
from typing import List, Tuple, Dict, Union, Any
# from finol.evaluation_layer.radar_chart import plot_radar_chart
from finol.utils import ROOT_PATH, load_config, add_prefix

# plt.style.use("seaborn-paper")
# plt.rcParams["font.family"] = "Microsoft YaHei"
# warnings.filterwarnings("ignore")
# warnings.simplefilter(action="ignore", category=FutureWarning)


class BenchmarkLoader:
    """
    Class to load the benchmarks and perform comparisons.
    """
    def __init__(self, caculate_metric_output: Dict, economic_distiller_caculate_metric_output: Dict) -> None:
        self.config = load_config()
        self.caculate_metric_output = caculate_metric_output
        self.economic_distiller_caculate_metric_output = economic_distiller_caculate_metric_output
        self.logdir = self.caculate_metric_output["logdir"]

    def find_top_5_baselines(self, df: pd.DataFrame) -> None:
        """
        Find the top 5 baselines based on the provided Benchmark DataFrame.

        :param df: Benchmark DataFrame containing all baselines.
        """
        df_copy = df.copy()
        df_copy = df_copy.drop(df.columns[-1], axis=1)
        df_copy.set_index('DATE', inplace=True)
        last_row = df_copy.iloc[-1]
        max_columns = last_row.nlargest(5).index.tolist()
        self.top_5_baselines = max_columns + [self.config["MODEL_NAME"]]
        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"]:
            self.top_5_baselines = self.top_5_baselines + [self.config["MODEL_NAME"] + ' (ED)']

    def load_benchmark(self) -> Dict[str, Any]:
        """
        Load benchmark data and update final profit results with model metrics.

        Returns: Dictionary containing benchmark results and model's results.
        """
        logdir = self.caculate_metric_output["logdir"]

        # ------------------------------------------------------------------------------------------------------------ #
        daily_return = pd.read_excel(ROOT_PATH + "/data/benchmark_results/profit_metrics/" + self.config["DATASET_NAME"] + "/daily_return.xlsx")
        daily_cumulative_wealth = pd.read_excel(ROOT_PATH + "/data/benchmark_results/profit_metrics/" + self.config["DATASET_NAME"] + "/daily_cumulative_wealth.xlsx")
        final_profit_result = pd.read_excel(ROOT_PATH + "/data/benchmark_results/profit_metrics/" + self.config["DATASET_NAME"] + "/final_profit_result.xlsx")

        daily_return = daily_return.dropna(axis=1, how="any")
        daily_cumulative_wealth = daily_cumulative_wealth.dropna(axis=1, how="any")
        final_profit_result = final_profit_result.dropna(axis=1, how="any")
        if self.caculate_metric_output is not None:
            daily_cumulative_wealth[self.config["MODEL_NAME"]] = self.caculate_metric_output["DCW"]
            final_profit_result = final_profit_result.assign(**{self.config["MODEL_NAME"]: np.nan})
            final_profit_result.loc[0, self.config["MODEL_NAME"]] = self.caculate_metric_output["CW"]
            final_profit_result.loc[1, self.config["MODEL_NAME"]] = self.caculate_metric_output["APY"]
            final_profit_result.loc[2, self.config["MODEL_NAME"]] = self.caculate_metric_output["SR"]
        if self.economic_distiller_caculate_metric_output is not None:
            daily_cumulative_wealth[self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["DCW"]
            final_profit_result = final_profit_result.assign(**{self.config["MODEL_NAME"] + " (ED)": np.nan})
            final_profit_result.loc[0, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["CW"]
            final_profit_result.loc[1, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["APY"]
            final_profit_result.loc[2, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["SR"]

        self.find_top_5_baselines(daily_cumulative_wealth)

        tabulate_data = []
        tabulate_data.append(["FCW"] + list(final_profit_result.loc[0, self.top_5_baselines].values))
        tabulate_data.append(["APY"] + list(final_profit_result.loc[1, self.top_5_baselines].values))
        tabulate_data.append(["SR"] + list(final_profit_result.loc[2, self.top_5_baselines].values))
        print("Profitability comparison with the top five baselines:")
        print(tabulate(tabulate_data, headers=["Profit Metric"] + list(final_profit_result.loc[0, self.top_5_baselines].index), tablefmt="psql", numalign="left"))

        # save results
        daily_cumulative_wealth.to_excel(logdir + "/" + add_prefix("daily_cumulative_wealth.xlsx"), index=False)
        final_profit_result.to_excel(logdir + "/" + add_prefix("final_profit_result.xlsx"), index=False)

        # ------------------------------------------------------------------------------------------------------------ #
        daily_drawdown = pd.read_excel(ROOT_PATH + "/data/benchmark_results/risk_metrics/" + self.config["DATASET_NAME"] + "/daily_drawdown.xlsx")
        final_risk_result = pd.read_excel(ROOT_PATH + "/data/benchmark_results/risk_metrics/" + self.config["DATASET_NAME"] + "/final_risk_result.xlsx")

        daily_drawdown = daily_drawdown.dropna(axis=1, how="any")
        final_risk_result = final_risk_result.dropna(axis=1, how="any")

        # calculate daily MDD
        daily_maximum_drawdown = daily_drawdown.copy()
        for col in daily_drawdown.columns:
            daily_maximum_drawdown[f"{col}"] = daily_drawdown[col].cummax()

        if self.caculate_metric_output != None:
            daily_drawdown[self.config["MODEL_NAME"]] = self.caculate_metric_output["DDD"]
            daily_maximum_drawdown[self.config["MODEL_NAME"]] = self.caculate_metric_output["DMDD"]
            final_risk_result = final_risk_result.assign(**{self.config["MODEL_NAME"]: np.nan})
            final_risk_result.loc[0, self.config["MODEL_NAME"]] = self.caculate_metric_output["VR"]
            final_risk_result.loc[1, self.config["MODEL_NAME"]] = self.caculate_metric_output["MDD"]
        if self.economic_distiller_caculate_metric_output != None:
            daily_drawdown[self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["DDD"]
            daily_maximum_drawdown[self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["DMDD"]
            final_risk_result = final_risk_result.assign(**{self.config["MODEL_NAME"] + " (ED)": np.nan})
            final_risk_result.loc[0, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["VR"]
            final_risk_result.loc[1, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["MDD"]

        tabulate_data = []
        tabulate_data.append(["VR"] + list(final_risk_result.loc[0, self.top_5_baselines].values))
        tabulate_data.append(["MDD"] + list(final_risk_result.loc[1, self.top_5_baselines].values))
        print("Risk resilience comparison with the top five baselines:")
        print(tabulate(tabulate_data, headers=["Risk Metric"] + list(final_risk_result.loc[0, self.top_5_baselines].index), tablefmt="psql", numalign="left"))

        # save results
        daily_drawdown.to_excel(logdir + "/" + add_prefix("daily_drawdown.xlsx"), index=False)
        daily_maximum_drawdown.to_excel(logdir + "/" + add_prefix("daily_maximumdrawdown.xlsx"), index=False)
        final_risk_result.to_excel(logdir + "/" + add_prefix("final_risk_result.xlsx"), index=False)

        # ------------------------------------------------------------------------------------------------------------ #
        transaction_costs_adjusted_cumulative_wealth = pd.read_excel(ROOT_PATH + "/data/benchmark_results/practical_metrics/" + self.config["DATASET_NAME"] + "/transaction_costs_adjusted_cumulative_wealth.xlsx")
        final_practical_result = pd.read_excel(ROOT_PATH + "/data/benchmark_results/practical_metrics/" + self.config["DATASET_NAME"] + "/final_practical_result.xlsx")

        transaction_costs_adjusted_cumulative_wealth = transaction_costs_adjusted_cumulative_wealth.dropna(axis=1, how="any")
        final_practical_result = final_practical_result.dropna(axis=1, how="any")
        if self.caculate_metric_output != None:
            transaction_costs_adjusted_cumulative_wealth[self.config["MODEL_NAME"]] = self.caculate_metric_output["TCW"]
            final_practical_result = final_practical_result.assign(**{self.config["MODEL_NAME"]: np.nan})
            final_practical_result.loc[0, self.config["MODEL_NAME"]] = self.caculate_metric_output["ATO"]
            final_practical_result.loc[1, self.config["MODEL_NAME"]] = self.caculate_metric_output["RT"]
        if self.economic_distiller_caculate_metric_output != None:
            transaction_costs_adjusted_cumulative_wealth[self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["TCW"]
            final_practical_result = final_practical_result.assign(**{self.config["MODEL_NAME"] + " (ED)": np.nan})
            final_practical_result.loc[0, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["ATO"]
            final_practical_result.loc[1, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["RT"]

        tabulate_data = []
        tabulate_data.append(["ATO"] + list(final_practical_result.loc[0, self.top_5_baselines].values))
        tabulate_data.append(["RT"] + list(final_practical_result.loc[1, self.top_5_baselines].values))
        print("Practicality comparison with the top five baselines:")
        print(tabulate(tabulate_data, headers=["Practical Metric"] + list(final_practical_result.loc[0, self.top_5_baselines].index), tablefmt="psql", numalign="left"))

        # save results
        transaction_costs_adjusted_cumulative_wealth.to_excel(logdir + "/" + add_prefix("transaction_costs_adjusted_cumulative_wealth.xlsx"), index=False)
        final_practical_result.to_excel(logdir + "/" + add_prefix("final_practical_result.xlsx"), index=False)

        # ------------------------------------------------------------------------------------------------------------ #
        # plot_radar_chart(final_profit_result, final_risk_result, self.COMPARED_BASELINE, logdir)

        load_benchmark_output = {}
        load_benchmark_output["logdir"] = self.caculate_metric_output["logdir"]
        load_benchmark_output["top_5_baselines"] = self.top_5_baselines
        load_benchmark_output["daily_cumulative_wealth"] = daily_cumulative_wealth
        load_benchmark_output["final_profit_result"] = final_profit_result
        load_benchmark_output["daily_drawdown"] = daily_drawdown
        load_benchmark_output["daily_maximum_drawdown"] = daily_maximum_drawdown
        load_benchmark_output["final_risk_result"] = final_risk_result
        load_benchmark_output["transaction_costs_adjusted_cumulative_wealth"] = transaction_costs_adjusted_cumulative_wealth
        load_benchmark_output["final_practical_result"] = final_practical_result
        # load_benchmark_output["CW"] = self.caculate_metric_output["CW"]
        # load_benchmark_output["TCW"] = self.caculate_metric_output["TCW"]
        # print(load_benchmark_output["TCW"])
        return load_benchmark_output


# if __name__ == "__main__":



import time
import torch
import optuna
import optuna.visualization.matplotlib as mpl
import matplotlib.pyplot as plt

from tqdm import tqdm
from tabulate import tabulate
from finol.model_layer.model_selector import ModelSelector
from finol.optimization_layer.criterion_selector import CriterionSelector
from finol.optimization_layer.optimizer_selector import OptimizerSelector
from finol.utils import load_config, update_config, portfolio_selection, set_seed


class ParametersTuner:
    """
    Class to tune model hyper-parameters using auto ML library.
    """
    def __init__(self, load_dataset_output) -> None:
        super().__init__()
        self.config = load_config()
        self.load_dataset_output = load_dataset_output
        self.train_loader = load_dataset_output["train_loader"]  # train_loader \ test_loader_for_train
        self.val_loader = load_dataset_output["val_loader"]  # val_loader \ test_loader_for_train
        self.test_loader = load_dataset_output["test_loader_for_train"]  # val_loader \ test_loader_for_train
        self.logdir = load_dataset_output["logdir"]

    def sample_params(self, trial: optuna.Trial) -> None:
        """
        Samples model hyper-parameters for optimization.

        :param trial: Optuna Trial object to sample the parameters.
        :return: None
        """
        model_sampled_params = self.config["MODEL_PARAMS_SPACE"][self.config["MODEL_NAME"]]
        self.sampled_params = {}

        for param_name, param_space in model_sampled_params.items():
            param_type = param_space["type"]
            param_range = param_space["range"]
            param_step = param_space["step"]

            if param_type == "int":
                self.sampled_params[param_name] = trial.suggest_int(param_name, *param_range, step=param_step)
            elif param_type == "float":
                self.sampled_params[param_name] = trial.suggest_float(param_name, *param_range, step=param_step)
            else:
                raise ValueError(f"Invalid parameter type: {param_type}")

    def objective(self, trial: optuna.Trial) -> float:
        """
        Objective function for Optuna optimization. Trains the model using the sampled parameters and returns the validation loss.

        :param trial: Optuna Trial object.
        :return: Validation loss.
        """
        self.sample_params(trial)
        set_seed(seed=self.config["MANUAL_SEED"])

        model = ModelSelector(self.load_dataset_output).select_model(self.sampled_params)
        optimizer = OptimizerSelector(model).select_optimizer()
        criterion = CriterionSelector()

        train_loss_list = []
        val_loss_list = []
        best_val_loss = float("inf")

        for e in tqdm(range(self.config["NUM_EPOCHES"]), desc="Training"):
            model.train()
            train_loss = 0
            for i, data in enumerate(self.train_loader, 1):
                x_data, label = data
                final_scores = model(x_data.float())
                portfolio = portfolio_selection(final_scores)

                loss = criterion(portfolio, label.float())

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                train_loss += loss.item()

            train_loss /= len(self.train_loader)
            train_loss_list.append(train_loss)

            if (e + 1) % self.config["SAVE_EVERY"] == 0:
                with torch.no_grad():
                    model.eval()
                    val_loss = 0
                    for i, data in enumerate(self.val_loader, 1):
                        val_data, label = data
                        final_scores = model(val_data.float())
                        portfolio = portfolio_selection(final_scores)
                        loss = criterion(portfolio, label.float())
                        val_loss += loss.item()

                    val_loss /= len(self.val_loader)
                    val_loss_list.append(val_loss)
                    # value = sum(val_loss_list) / len(val_loss_list)
                    value = min(val_loss_list)  # use this
                    # value = val_loss  # real time val loss

                    if val_loss < best_val_loss:
                        best_val_loss = val_loss

            # Report intermediate objective value.
            trial.report(value, e)

            # Handle pruning based on the intermediate value.
            if trial.should_prune():
                raise optuna.TrialPruned()
        return value

    def select_sampler(self) -> object:
        """
        Selects and initializes an Optuna sampler based on the configuration.

        :return: Initialized Optuna sampler object.
        """
        sampler_name = self.config["SAMPLER_NAME"]
        sampler_kwargs = {"seed": self.config["MANUAL_SEED"]}

        if sampler_name == "GridSampler":
            model_sampled_params = self.config["MODEL_PARAMS_SPACE"][self.config["MODEL_NAME"]]
            search_space = {}
            for param_name, param_space in model_sampled_params.items():
                param_type = param_space["type"]
                param_range = param_space["range"]
                if param_type == "int" or param_type == "float":
                    search_space[param_name] = param_range
            sampler_kwargs = {"search_space": search_space}

        return getattr(optuna.samplers, sampler_name)(**sampler_kwargs)

    def select_pruner(self) -> object:
        """
        Selects and initializes an Optuna pruner based on the configuration.

        :return: Initialized Optuna pruner object.
        """
        pruner_name = self.config["PRUNER_NAME"]
        pruner_kwargs = {}

        if pruner_name == "PatientPruner":
            pruner_kwargs["wrapped_pruner"] = getattr(optuna.pruners, self.config["WRAPPED_PRUNER_NAME"])()
            pruner_kwargs["patience"] = 1

        return getattr(optuna.pruners, pruner_name)(**pruner_kwargs)

    def tune_parameters(self) -> None:
        """
        Tune model hyper-parameters.

        This method creates an Optuna study object, optimizes the objective function, visualizes the results, and prints out the optimization results.

        :return: None
        """
        # Creating Optuna object and defining its parameters
        self.study = optuna.create_study(
            direction="minimize",
            sampler=self.select_sampler(),
            pruner=self.select_pruner(),
            storage="sqlite:///" + self.logdir + "/" + self.config["MODEL_NAME"] + "_" + self.config["DATASET_NAME"] + "_" + self.config["MODEL_NAME"] + ".db"
        )
        self.study.optimize(self.objective, n_trials=self.config["NUM_TRIALS"])

        # To further visualize the results, you can upload the generated {MODEL_NAME}.db file to the Optuna Dashboard:
        # https://optuna.github.io/optuna-dashboard/
        # optuna.visualization.plot_optimization_history(self.study).show()
        plots = [
            mpl.plot_contour,
            mpl.plot_edf,
            # mpl.plot_hypervolume_history,
            mpl.plot_intermediate_values,
            mpl.plot_optimization_history,
            mpl.plot_parallel_coordinate,
            mpl.plot_param_importances,
            # mpl.plot_pareto_front,
            mpl.plot_rank,
            mpl.plot_slice,
            mpl.plot_terminator_improvement,
            mpl.plot_timeline,
        ]
        for plot_func in plots:
            fig = plot_func(self.study)
            plt.show()
            # plt.tight_layout()

        # Showing optimization results
        tabulate_data = [
            ["Sampler", self.config["SAMPLER_NAME"]],
            ["Pruner", self.config["PRUNER_NAME"]],
            ["Number of finished trials", len(self.study.trials)],
            ["Best trial parameters", self.study.best_trial.params],
            ["Best score", self.study.best_value]
        ]
        print(tabulate(tabulate_data, headers=["Auto Hyper-parameters Tuning", "INFO"], tablefmt="psql"))  # , headers=["Metric", "Value"]

        # Write config
        self.config["MODEL_PARAMS"][self.config["MODEL_NAME"]] = self.study.best_trial.params
        update_config(self.config)


if __name__ == "__main__":
    from finol.data_layer.dataset_loader import DatasetLoader
    load_dataset_output = DatasetLoader().load_dataset()
    ParametersTuner(load_dataset_output=load_dataset_output).tune_parameters()



import time

import torch

from finol.utils import load_config


class CriterionSelector:
    """
    Class to select and compute different loss criteria for portfolio selection.
    """
    def __init__(self) -> None:
        self.config = load_config()
        self.criterion_dict = {
            "LogWealth": self.compute_log_wealth_loss,
            "LogWealthL2Diversification": self.compute_log_wealth_l2_diversification_loss,
            "LogWealthL2Concentration": self.compute_log_wealth_l2_concentration_loss,
            "L2Diversification": self.compute_l2_diversification_loss,
            "L2Concentration": self.compute_l2_concentration_loss,
            "SharpeRatio": self.compute_sharpe_ratio_loss,
            "Volatility": self.compute_volatility_loss,
        }

    def compute_log_wealth_loss(self, portfolios: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        """
        Compute the ``LogWealth`` loss, which calculates the negative mean logarithm of the portfolio's daily returns.

        :param portfolios: Portfolio weights tensor of shape (batch_size, num_assets).
        :param labels: Label tensor representing asset returns of shape (batch_size, num_assets).
        :return: ``LogWealth`` loss tensor, representing the negative mean logarithm of the portfolio's daily returns.
        """
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        log_returns = torch.log(torch.clamp(daily_returns, min=1e-6))
        loss = - torch.mean(log_returns)
        return loss

    def compute_log_wealth_l2_diversification_loss(self, portfolios: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        """
        Compute the ``LogWealthL2Diversification`` loss, which extends the ``LogWealth`` loss by incorporating diversification measures.

        This loss function calculates the negative mean logarithm of the portfolio's daily returns while considering diversification metrics.

        :param portfolios: Portfolio weights tensor of shape (batch_size, num_assets).
        :param labels: Label tensor representing asset returns of shape (batch_size, num_assets).
        :return: ``LogWealthL2Diversification`` loss tensor, a modified version of ``LogWealth`` loss with diversification metrics included.
        """
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        log_returns = torch.log(torch.clamp(daily_returns, min=1e-6))

        # Calculate the similarity loss
        # We use the L2 norm to measure the difference between consecutive predictions
        diff = portfolios[1:, :] - portfolios[:-1, :]
        diff_loss = torch.norm(diff, p=2, dim=-1)

        # By minimizing the L2 norm of the difference vectors, you will tend to get two vectors that are similar, since the
        # L2 norm measures the Euclidean distance between vectors. When the L2 norm is small, it means that the difference
        # between two vectors is small.
        loss = - torch.mean(log_returns) + self.config["LAMBDA_L2"] * torch.mean(diff_loss)
        return loss

    def compute_log_wealth_l2_concentration_loss(self, portfolios: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        """
        Compute the ``LogWealthL2Concentration`` loss, which extends the ``LogWealth`` loss by incorporating concentration measures.

        :param portfolios: Portfolio weights tensor of shape (batch_size, num_assets).
        :param labels: Label tensor representing asset returns of shape (batch_size, num_assets).
        :return: ``LogWealthL2Concentration`` loss tensor, a modified version of ``LogWealth`` loss with concentration metrics included.
        """
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        log_returns = torch.log(torch.clamp(daily_returns, min=1e-6))

        # Calculate the similarity loss
        # We use the L2 norm to measure the difference between consecutive predictions
        diff = portfolios[1:, :] - portfolios[:-1, :]
        diff_loss = torch.norm(diff, p=2, dim=-1)

        # By minimizing the L2 norm of the difference vectors, you will tend to get two vectors that are similar, since the
        # L2 norm measures the Euclidean distance between vectors. When the L2 norm is small, it means that the difference
        # between two vectors is small.
        loss = - torch.mean(log_returns) - self.config["LAMBDA_L2"] * torch.mean(diff_loss)
        return loss

    def compute_l2_diversification_loss(self, portfolios: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        """
        Compute the ``L2Diversification`` loss, which measures the diversification of a portfolio based on the L2 norm of consecutive portfolio weight differences.

        This loss function calculates the mean L2 norm of the differences between consecutive portfolio weight vectors.

        :param portfolios: Portfolio weights tensor of shape (batch_size, num_assets).
        :param labels: Label tensor representing asset returns of shape (batch_size, num_assets).
        :return: ``L2Diversification`` loss tensor, indicating the degree of diversification based on the L2 norm of portfolio weight differences.
        """
        # Calculate the similarity loss
        # We use the L2 norm to measure the difference between consecutive predictions
        diff = portfolios[1:, :] - portfolios[:-1, :]
        diff_loss = torch.norm(diff, p=2, dim=-1)

        # By minimizing the L2 norm of the difference vectors, you will tend to get two vectors that are similar, since the
        # L2 norm measures the Euclidean distance between vectors. When the L2 norm is small, it means that the difference
        # between two vectors is small.
        loss = torch.mean(diff_loss)
        return loss

    def compute_l2_concentration_loss(self, portfolios: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        """
        Compute the ``L2Concentration`` loss, which measures the concentration of a portfolio based on the L2 norm of consecutive portfolio weight differences.

        This loss function calculates the negative mean L2 norm of the differences between consecutive portfolio weight vectors.

        :param portfolios: Portfolio weights tensor of shape (batch_size, num_assets).
        :param labels: Label tensor representing asset returns of shape (batch_size, num_assets).
        :return: ``L2Concentration`` loss tensor, indicating the degree of concentration based on the negative mean L2 norm of portfolio weight differences.
        """
        # Calculate the similarity loss
        # We use the L2 norm to measure the difference between consecutive predictions
        diff = portfolios[1:, :] - portfolios[:-1, :]
        diff_loss = torch.norm(diff, p=2, dim=-1)

        # By minimizing the L2 norm of the difference vectors, you will tend to get two vectors that are similar, since the
        # L2 norm measures the Euclidean distance between vectors. When the L2 norm is small, it means that the difference
        # between two vectors is small.
        loss = - torch.mean(diff_loss)
        return loss

    def compute_sharpe_ratio_loss(self, portfolios: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        """
        Compute the ``SharpeRatio`` loss, which evaluates the risk-adjusted return of the portfolios using the Sharpe Ratio.

        This loss function calculates the negative Sharpe Ratio, which is the ratio of the mean excess return to the standard deviation of the excess return.

        :param portfolios: Portfolio weights tensor of shape (batch_size, num_assets).
        :param labels: Label tensor representing asset returns of shape (batch_size, num_assets).
        :return: ``SharpeRatio`` loss tensor, indicating the negative Sharpe Ratio value for the portfolios.
        """
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        # log_returns = torch.log(torch.clamp(daily_returns, min=1e-6))
        excess_returns = daily_returns  # risk_free_rate = 0

        mean_excess_return = torch.mean(excess_returns)
        std_excess_return = torch.std(excess_returns)

        sharpe_ratio = mean_excess_return / (std_excess_return + 1e-6)
        loss = - sharpe_ratio
        return loss

    def compute_volatility_loss(self, portfolios: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        """
        Compute the ``Volatility`` loss, which measures the volatility of the portfolios based on the standard deviation of daily returns.

        This loss function calculates the standard deviation of the daily portfolio returns.

        :param portfolios: Portfolio weights tensor of shape (batch_size, num_assets).
        :param labels: Label tensor representing asset returns of shape (batch_size, num_assets).
        :return: ``Volatility`` loss tensor, representing the volatility of the portfolios based on the standard deviation of daily returns.
        """
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        volatility = torch.std(daily_returns)
        loss = volatility
        return loss

    def __call__(self, portfolios: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        criterion_cls = self.criterion_dict.get(self.config["CRITERION_NAME"], None)
        if criterion_cls is None:
            raise ValueError(f"Invalid criterion name: {self.config['CRITERION_NAME']}. Supported criterions are: {self.criterion_dict.keys()}")
        return criterion_cls(portfolios, labels)



from rich import print
from finol.data_layer.dataset_loader import DatasetLoader
from finol.data_layer.scaler_selector import ScalerSelector
from finol.utils import load_config, update_config, detect_device

__all__ = [
    "DatasetLoader",
    "ScalerSelector",
]

config = load_config()
detect_device(config)
update_config(config)


import time
import torch

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from IPython import display
from tqdm import tqdm
from typing import List, Tuple, Dict, Union
from finol.model_layer.model_selector import ModelSelector
from finol.optimization_layer.criterion_selector import CriterionSelector
from finol.optimization_layer.optimizer_selector import OptimizerSelector
from finol.optimization_layer.parameters_tuner import ParametersTuner
from finol.utils import load_config, portfolio_selection, set_seed, add_prefix


class ModelTrainer:
    """
    Class to train a machine learning model for portfolio selection.
    """
    def __init__(self, load_dataset_output) -> None:
        self.config = load_config()
        self.load_dataset_output = load_dataset_output
        self.logdir = load_dataset_output["logdir"]
        self.train_loader = load_dataset_output["train_loader"]  # train_loader \ test_loader_for_train
        self.val_loader = load_dataset_output["val_loader"]  # val_loader \ test_loader_for_train
        self.test_loader = load_dataset_output["test_loader"]  # test_loader \ test_loader_for_train

        self.is_ipython = "inline" in matplotlib.get_backend()
        # print(self.is_ipython)

    def plot_loss_notebook(self) -> None:
        """
        Plot the training and validation losses in a notebook environment.

        This method plots the average training and validation losses over epochs using matplotlib in an interactive notebook environment.

        :return: None
        """
        if self.is_ipython:
            plt.ion()
            plt.figure()  # figsize=(12, 5)
            plt.clf()
            plt.plot(self.avg_train_loss_list, linestyle="-", marker=self.config["MARKERS"][0], markevery=int(self.config["NUM_EPOCHES"]/20), color="black", alpha=0.5, label="train loss")
            plt.plot(self.avg_val_loss_list, linestyle=":", marker=self.config["MARKERS"][1], markevery=int(self.config["NUM_EPOCHES"]/20), color="black", alpha=0.5, label="val loss")
            plt.xlabel("Epoch")
            plt.ylabel("Loss")
            plt.legend()
            plt.grid(True)
            display.clear_output(wait=True)
            display.display(plt.gcf())
            plt.tight_layout()
            # plt.yscale("log")
            plt.savefig(self.logdir + "/" + add_prefix("loss.pdf"), format="pdf", dpi=300, bbox_inches="tight")
            plt.clf()
            plt.close()

    def plot_loss(self) -> None:
        """
        Plot the training and validation losses.

        This method plots the average training and validation losses over epochs using matplotlib.

        :return: None
        """
        # not_ipython = "inline" not in matplotlib.get_backend()
        if not self.is_ipython:
            plt.figure()  # figsize=(12, 5)
            plt.plot(np.array(self.avg_train_loss_list), linestyle="-", marker=self.config["MARKERS"][0], color="black", alpha=0.5, label="train loss")
            plt.plot(np.array(self.avg_val_loss_list), linestyle=":", marker=self.config["MARKERS"][1], color="black", alpha=0.5, label="val loss")
            # markevery = int(self.config["NUM_EPOCHES"] / 20)
            plt.xlabel("Epoch")
            plt.ylabel("Loss")
            plt.legend()
            plt.grid(True)
            plt.tight_layout()
            # plt.yscale("log")
            plt.savefig(self.logdir + "/" + add_prefix("loss.pdf"), format="pdf", dpi=300, bbox_inches="tight")
            plt.show()

    def train_model(self) -> Dict:
        """
        Train the machine learning model using the specified dataset.

        This method iterates over the specified number of epochs and trains the model using the training data.
        It calculates and stores the training and validation losses, saves the best model based on validation loss,
        and optionally plots the dynamic loss during training.

        :return: Dictionary containing the log directory for the trained model.
        """
        if self.config["TUNE_PARAMETERS"]:
            ParametersTuner(load_dataset_output=self.load_dataset_output).tune_parameters()
            pass

        set_seed(seed=self.config["MANUAL_SEED"])
        model = ModelSelector(self.load_dataset_output).select_model()
        optimizer = OptimizerSelector(model).select_optimizer()
        criterion = CriterionSelector()

        self.train_loss_list = []
        self.val_loss_list = []
        self.avg_train_loss_list = []
        self.avg_val_loss_list = []
        best_val_loss = float("inf")

        for e in tqdm(range(self.config["NUM_EPOCHES"]), desc="Training"):
            model.train()
            train_loss = 0
            for i, data in enumerate(self.train_loader, 1):
                x_data, label = data
                final_scores = model(x_data.float())
                portfolio = portfolio_selection(final_scores)
                loss = criterion(portfolio, label.float())

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                train_loss += loss.item()

            train_loss /= len(self.train_loader)
            self.train_loss_list.append(train_loss)
            self.avg_train_loss_list.append(sum(self.train_loss_list) / len(self.train_loss_list))

            if (e + 1) % self.config["SAVE_EVERY"] == 0:
                with torch.no_grad():
                    model.eval()
                    val_loss = 0
                    for i, data in enumerate(self.val_loader, 1):
                        val_data, label = data
                        final_scores = model(val_data.float())
                        portfolio = portfolio_selection(final_scores)
                        loss = criterion(portfolio, label.float())
                        val_loss += loss.item()

                    val_loss /= len(self.val_loader)
                    self.val_loss_list.append(val_loss)
                    self.avg_val_loss_list.append(sum(self.val_loss_list) / len(self.val_loss_list))

                    if val_loss < best_val_loss:
                        best_val_loss = val_loss
                        # best_model = model
                        # print("best_model", "e:", e)
                        torch.save(model, self.logdir + "/" + add_prefix("best_model.pt"))

            if self.config["PLOT_DYNAMIC_LOSS"]:
                if (e + 1) % 10 == 0:
                    # print("Epoch: {}, Train Loss: {}, Val Loss: {}".format(e + 1, train_loss, val_loss))
                    self.plot_loss_notebook()

        self.plot_loss_notebook()
        self.plot_loss()

        train_model_output = {
            "logdir": self.logdir,
            # "last_model": model,
            # "best_model": best_model
        }
        # print(
        #     best_model
        # )
        torch.save(model, self.logdir + "/" + add_prefix("last_model.pt"))
        return train_model_output


if __name__ == "__main__":
    from finol.data_layer.dataset_loader import DatasetLoader
    from finol.evaluation_layer.model_evaluator import ModelEvaluator
    load_dataset_output = DatasetLoader().load_dataset()
    train_model_output = ModelTrainer(load_dataset_output).train_model()
    evaluate_model_output = ModelEvaluator(load_dataset_output, train_model_output).evaluate_model()